#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <time.h>
#include <fcntl.h>
#include <errno.h>

#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>

/*************************************************

  Module for Antmon system. Gathers information from
  SNMP sources

  Format of sensor names: hostname:OID
  E.g. 192.168.0.1:.1.3.6.1.2.1.2.2.1.12.2

**************************************************/


#define REQ_BUFFER 40240
#define PATH_LEN 4096
#define DATA_LEN 4096
#define USECS_MIN 1000000
#define BUF_SIZE 40960
#define SLEEP_USECS 1000000
#define SMALL_VALUE 0.1


#define debug 0

/* global variables */
char req[REQ_BUFFER+1];
char base[PATH_LEN]="/";
struct timeval global_time;



/* async read/write */
void print(const char *s,...);
char *getstr(int fd);

/* process one request (prepare data only)*/
void process_line(const char *s);
/* print available sensors anmes */
void print_sensors(void);
/* reset sensors data */
/*void reset_data(void);*/

/* do actual requests and pribt data */
void do_request(void);

void reset_data(void);


oid *num2oid(const char *str,
                oid *root,
                size_t *rootlen);

/*****************************************
 *
 *
 *     Store data and data rates
 *
 *     No deleting is implemented
 *
 *
*****************************************/

struct float_data{
    char              *name;
    float             data;
    float             rate;
    struct timeval    tv;
    struct float_data *next;
};

struct float_data float_list={"",0.0,0.0,{0,0},NULL};

struct float_data *new_float_data(const char *name){
    struct float_data *ptr;

    if(name==NULL)
        return NULL;
    ptr=&float_list;
    while(ptr->next!=NULL){
        ptr=ptr->next;
    }
    ptr->next=malloc(sizeof(struct float_data));
    ptr=ptr->next;
    ptr->next=NULL;
    ptr->name=malloc(strlen(name)+1);
    strcpy(ptr->name,name);
    return ptr;
}

struct float_data *get_float_data(const char *name){
    struct float_data *ptr;

    if(name==NULL)
        return NULL;
    ptr=float_list.next;
    while(ptr!=NULL){
        if(!strcmp(ptr->name,name)){
            return ptr;
        }
        ptr=ptr->next;
    }
    return NULL;
}

void clean_float_data(void){
    struct float_data *ptr,*curr;

    ptr=float_list.next;
    while(ptr!=NULL){
        curr=ptr;
        ptr=ptr->next;
        free(curr->name);
        free(curr);
    }
}

struct float_data *first_float_data(void){
    return float_list.next;
}

struct float_data *next_float_data(struct float_data *f){
    return f->next;
}
/****************************************
 *
 * sessions list
 *
 ****************************************/

struct sess_data{
    char                *name;
    struct snmp_session *sess;
    struct snmp_pdu     *pdu;
    struct timeval      tv;
    struct sess_data    *next;
};

struct sess_data sess_list={"",NULL,NULL,{0,0},NULL};

struct sess_data *new_sess_data(const char *name){
    struct sess_data *ptr;

    if(name==NULL)
        return NULL;
    ptr=&sess_list;
    while(ptr->next!=NULL){
        ptr=ptr->next;
    }
    ptr->next=malloc(sizeof(struct sess_data));
    ptr=ptr->next;
    ptr->next=NULL;
    ptr->name=malloc(strlen(name)+1);
    strcpy(ptr->name,name);
    return ptr;
}

struct sess_data *get_sess_data(const char *name){
    struct sess_data *ptr;

    if(name==NULL)
        return NULL;
    ptr=sess_list.next;
    while(ptr!=NULL){
        if(!strcmp(ptr->name,name)){
            return ptr;
        }
        ptr=ptr->next;
    }
    return NULL;
}

void clean_sess_data(void){
    struct sess_data *ptr,*curr;

    ptr=sess_list.next;
    while(ptr!=NULL){
        curr=ptr;
        ptr=ptr->next;
        free(curr->name);
        free(curr);
    }
    sess_list.next=NULL;
}

struct sess_data *first_sess_data(void){
    return sess_list.next;
}

struct sess_data *next_sess_data(struct sess_data *f){
    return f->next;
}

/****************************************
 *
 * requests names list
 *
 ****************************************/

struct req_data{
    char                *name;
    float               data;
    int                 got;
    struct req_data     *next;
};

struct req_data req_list={"",0.0,0,NULL};

struct req_data *new_req_data(const char *name){
    struct req_data *ptr;

    if(name==NULL)
        return NULL;
    ptr=&req_list;
    while(ptr->next!=NULL){
        ptr=ptr->next;
    }
    ptr->next=malloc(sizeof(struct req_data));
    ptr=ptr->next;
    ptr->next=NULL;
    ptr->data=0.0;
    ptr->got=0;
    ptr->name=malloc(strlen(name)+1);
    strcpy(ptr->name,name);
    return ptr;
}

struct req_data *get_req_data(const char *name){
    struct req_data *ptr;

    if(name==NULL)
        return NULL;
    ptr=req_list.next;
    while(ptr!=NULL){
        if(!strcmp(ptr->name,name)){
            return ptr;
        }
        ptr=ptr->next;
    }
    return NULL;
}

void clean_req_data(void){
    struct req_data *ptr,*curr;

    ptr=req_list.next;
    while(ptr!=NULL){
        curr=ptr;
        ptr=ptr->next;
        free(curr->name);
        free(curr);
    }
    req_list.next=NULL;
}

struct req_data *first_req_data(void){
    return req_list.next;
}

struct req_data *next_req_data(struct req_data *f){
    return f->next;
}

/*****************************************
 *
 *
 *   main loop
 *
 *
 *****************************************/
int main(int argc, char **argv){
    char p,p2;
    int  i,index,r;

    /* check for argument */
    if(argc>1){
        strncpy(base,argv[1],PATH_LEN);
    }

    /* important !!! */
    /* make i/o nonblocked */
    fcntl(0,F_SETFL,fcntl(0,F_GETFL,0)|O_NONBLOCK);
    fcntl(1,F_SETFL,fcntl(1,F_GETFL,0)|O_NONBLOCK);

    /* first read = init */
    for(;;){
        if(read(0,&p,1)<1){
            usleep(SLEEP_USECS);
            continue;
        }
        if(p=='\n')
            break;
    }


    setenv("MIBS","all",1);
    init_snmp("AntMon");
    init_mib();
    /*read_module("ALL");*/
    read_all_mibs();

    print_sensors();
/*    reset_data();*/

    /* main loop */
    for(;;){

        i=0;
        if(debug){
            fprintf(stderr,"---------------\n");
        }

        /* reset readed chars!!! */
        p=p2='\0';

        /* one-by-one byte */
        for(;;){
            r=read(0,&p,1);
            if(r==0){ /* eof */
                clean_float_data();
                return 0;
            }
            if(r>0){
/*
                if(debug)
                    fprintf(stderr,"GOT[%d]: %c\n",i,p);
*/
              if(p == '\n'){
                    if(p2 == '\n'){
                        req[i]='\0';
                        break;
                    }
                }
                req[i++]=p;
                if(i>REQ_BUFFER){
                    fprintf(stderr,"Request too long.\n");
                    exit(2);
                }
                if((p == '\n') && (p2 == '\n')){
                    break;
                }
                p2=p;
            }
            else
                usleep(SLEEP_USECS);
        }

        if(debug){
            fprintf(stderr,"REQ: '%s'\n",req);
        }

        gettimeofday(&global_time,NULL);
        /* process request ! */
        index=0;
        for(i=0;req[i]!='\0';){

            /* line-by-line */
            while(req[i]!='\n')
                ++i;
            req[i]='\0';
            ++i;
            if(debug){
                fprintf(stderr,"PROCESS: '%s'\n",req+index);
            }
            /* remember request name */
            new_req_data(req+index);
            /* process request */
            process_line(req+index);
            index=i;
        }
        do_request();
        clean_req_data();
        clean_sess_data();
        /* final empty line */
        print("\n");
    }/* main loop */

    clean_float_data();
    return 0;
}

/***********************************************
 * ********************************************/

/* print string in async mode. */
void print(const char *s,...){
  static char buffer[BUF_SIZE];
  int len,written,shift;
  va_list list;

  if(s==NULL)
    return;

  va_start(list,s);
  vsnprintf(buffer,BUF_SIZE-1,s,list);
  va_end(list);

  buffer[BUF_SIZE-1]='\0';

  if(debug){
    if(strcmp(buffer,"\n")) /* do not log only eol */
      fprintf(stderr,"SS-> '%s'\n",buffer);
  }

  len=strlen(buffer);
  shift=0;
  while(shift<len){
    written=write(0,buffer+shift,len-shift);

    /* was there an error? */
    if(written<0){
      if((errno==EAGAIN)||(errno==EINTR)){

        /* just sleep and try again */
        usleep(100);
      }
      else{
        exit(1);
      }
    }
    else{ /* all ok */
      shift+=written;
    }
  }
}

/* reads a string in asyncronous mode */
char *getstr(int fd){
  static char buffer[BUF_SIZE];

  int shift=0,res;
  char c;
  for(;;){
    res=read(fd,&c,1);
    if(res==0){ /* eof */
      if(shift>0)
        break;
      /* eof at first character - error! */
      return NULL;
    }
    if((res<0) && (errno==EAGAIN)){
      /* just sleep and try again */
      usleep(100);
      continue;
    }

    if(res<0){ /* error or eof */
      if(shift)
        break;
      return NULL;
    }

    /* add to buffer */
    buffer[shift]=c;
    if((c=='\0') || (c=='\n'))
      break;
    if(++shift>BUF_SIZE-2)
      break;
  }
  buffer[shift]='\0';
  return buffer;
}
/********************************************************
 ********************************************************
 ********************************************************
 ********************************************************
 ********************************************************/

/**************************************************
 *
 *
 * Real data harvesting section
 *
 *
 **************************************************/

/* process one request (prepare data only) */
void process_line(const char *s){
    struct snmp_session *sess;
    struct sess_data *sp;
    char host[PATH_LEN];
    char oid_name[PATH_LEN];
    oid id_oid[MAX_OID_LEN];
    oid *resolved_oid;
    size_t id_len;
    oid serial_oid[MAX_OID_LEN];
    int  i,fd,rate;
    char data[DATA_LEN], *ptr;
    struct float_data *fp;

    /* find hostname */
    i=0;
    while((s[i]!='\0') && (s[i]!=':'))
        ++i;
    if((i>=PATH_LEN) || (i==0)){
        fprintf(stderr, "Bad request '%s'\n",s);
        /*print("n/a\n");*/
        return;
    }
    memcpy(host,s,i);
    host[i]='\0';

    /* create new session record if needed */
    sp=get_sess_data(host);
    if(sp==NULL){
        sess=malloc(sizeof(struct snmp_session));
        snmp_sess_init( sess );
        sess->version = SNMP_VERSION_1;
        sess->community = "public";
        sess->community_len = strlen(sess->community);
        sess->peername = malloc(strlen(host)+1);
        strcpy(sess->peername,host);
        sp=new_sess_data(host);
        sp->sess=sess;
        sp->pdu=NULL;
    }

    if(sp->pdu==NULL){
        sp->pdu=snmp_pdu_create(SNMP_MSG_GET);
    }

    /* find OID */
    if(s[i]!='\0'){
        sscanf(s+i+1,"%s",oid_name);
    }
    else{
        fprintf(stderr, "Bad request '%s'\n",s);
/*        print("n/a\n");*/
        return;
    }

#ifdef OLDOLDOLD
    fp=get_float_data(s);

    if(fp!=NULL){
        /* was data just received? */
        if((global_time.tv_sec-fp->tv.tv_sec)*1000000+
           (global_time.tv_usec-fp->tv.tv_usec)<USECS_MIN){
            /* do not update anything... */
            return;
        }
    }
#endif

    /*if(read_objid(oid_name, id_oid, &id_len)!=1){*/
    if((resolved_oid=snmp_parse_oid(oid_name, id_oid, &id_len))==NULL){
    /*if(num2oid(oid_name,id_oid,&id_len)==NULL){*/
        fprintf(stderr, "Cannot resolve '%s'\n",oid_name);
        return;
    }
    snmp_add_null_var(sp->pdu, id_oid, id_len);
} /* process_line */
/**
  struct snmp_pdu *pdu;
  struct snmp_pdu *response;
  struct variable_list *vars;
  oid id_oid[MAX_OID_LEN];
  oid serial_oid[MAX_OID_LEN];
  size_t id_len = MAX_OID_LEN;
  size_t serial_len = MAX_OID_LEN;
  int status;
  struct tree * mib_tree;

  sess_handle = snmp_open(&session);
  add_mibdir(".");
  mib_tree = read_mib("PowerNet-MIB.txt");
  pdu = snmp_pdu_create(SNMP_MSG_GET);
  read_objid("PowerNet-MIB::upsBasicIdentModel.0", id_oid, &id_len);
   snmp_add_null_var(pdu, id_oid, id_len);
  read_objid("PowerNet-MIB::upsAdvIdentSerialNumber.0", serial_oid, &serial_len);
   snmp_add_null_var(pdu, serial_oid, serial_len);

 */

/*
ASN_BOOLEAN
ASN_INTEGER
ASN_BIT_STR
ASN_OCTET_STR
ASN_NULL
ASN_OBJECT_ID
ASN_SEQUENCE
ASN_SET
*/

/****************************************
 *
 *  Do actual SNMP requests
 *
 *
 ****************************************/
void do_request(){
    struct sess_data *sp;
    struct req_data *rptr;

    struct snmp_session *sh;
    struct snmp_pdu *response;
    struct variable_list *vars;
    int status;
    int j;
    char tmpbuf[BUF_SIZE];

    char sensor[BUF_SIZE];

    sp=first_sess_data();
    while(sp!=NULL){

        sh = snmp_open(sp->sess);

        status = snmp_synch_response(sh, sp->pdu, &response);
        if(status!=0){
                snmp_free_pdu(response);
                sp->pdu=NULL;
                snmp_close(sh);

                sp=next_sess_data(sp);
                continue;
        }
        for(vars = response->variables; vars; vars = vars->next_variable){
            /* host:sensor */
            sprintf(sensor,"%s:",sp->name);
            for(j=0;j<vars->name_length;++j){
                sprintf(tmpbuf,".%d",vars->name[j]);
                strcat(sensor,tmpbuf);
            }

            if((rptr=get_req_data(sensor))==NULL){
                break;
            }

            switch(vars->type) {
            /* int */
            case ASN_INTEGER:
                rptr->data=(float)*(vars->val.integer);
                rptr->got=1;
                break;
            case ASN_UNSIGNED:
            case ASN_TIMETICKS:
            case ASN_COUNTER:
                rptr->data=(float)(unsigned long)*(vars->val.integer);
                rptr->got=1;
                break;
            case ASN_COUNTER64:
                rptr->data=(float)(unsigned long)(vars->val.counter64->low);
                rptr->data+=
                    ((float)(unsigned long)(vars->val.counter64->high))*4294967296;
                rptr->got=1;
                break;

            /* string */
            case ASN_OCTET_STR:
            case ASN_BIT_STR:
            case ASN_OPAQUE:
            case ASN_NSAP:
                fprintf(stderr, "%s: wrong data type (str)\n", sensor);
                break;
            /* bool */
            case ASN_BOOLEAN:
                fprintf(stderr, "%s: wrong data type (bool)\n", sensor);
                break;
            /* what??? */
            default:
                fprintf(stderr, "%s: wrong data type (unknown)\n", sensor);
            }
        }

        /*snmp_free_pdu(response);*/
        sp->pdu=NULL;
        snmp_close(sh);

        sp=next_sess_data(sp);
    }

    rptr=first_req_data();
    while(rptr!=NULL){
        if(rptr->got==0){
            print("n/a\n");
        }
        else{
            print("%f\n",rptr->data);
        }
        rptr=next_req_data(rptr);
    }
} /* do_request */


/* print available sensors names */
void print_sensors(void){

    print("*\n\n");
    return;
} /*print_sensors */

/* reset sensors data
void reset_data(void){
    struct float_data *fd=first_float_data();

    gettimeofday(&global_time,NULL);

    while(fd!=NULL){
        fd->data=0.0;
        fd->rate=0.0;
        fd->tv.tv_sec=global_time.tv_sec;
        fd->tv.tv_usec=global_time.tv_usec;
        process_line(fd->name);

        fd=next_float_data(fd);
    }
} /* reset_data */

oid *snmp_parse_oid(const char *str,
                oid *root,
                size_t *rootlen)
{
  size_t savlen = *rootlen;
  if (read_objid(str,root,rootlen)) {
    return root;
  }
  *rootlen = savlen;
  fprintf(stderr,"parse_oid: get_node '%s'\n",str); /**/
  if (get_node(str,root,rootlen)) {
    return root;
  }
  *rootlen = savlen;
  fprintf(stderr,"parse_oid: wildly parsing '%s'\n",str); /**/
  if (get_wild_node(str,root,rootlen)) {
    return root;
  }
  return NULL;
}

/*************************************
 *
 *  Convert NUMERICAL string (.1.2.3.4.5.6...)
 *   into OID array
 *
 *************************************/

oid *num2oid(const char *str,
                oid *root,
                size_t *rootlen)
{
  size_t savlen = *rootlen;
  *rootlen=0;

  if(*str!='.')
    return NULL;

  ++str;
  while(*str!='\0'){
    *root=0;
    while(*str>='0' && *str<='9'){
        *root *=10;
        *root +=*str-'0';
        ++str;
    }
    if(*str=='.' || *str=='\0'){
        ++*rootlen;
        ++root;
        if(*str=='\0')
            return root;
        ++str;
        continue;
    }
    break;
  }
  *rootlen = savlen;
  return NULL;
}
